#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it 
#   under the terms of the GNU Lesser General Public License as published 
#   by the Free Software Foundation, either version 3 of the License, or 
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but 
#   WITHOUT ANY WARRANTY; without even the implied warranty of 
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public 
#   License along with Dragonfly.  If not, see 
#   <http://www.gnu.org/licenses/>.
#

"""
Window class
============================================================================

"""
from six import string_types, integer_types

import win32gui
import win32con
from ctypes          import windll, pointer, c_wchar, c_ulong

from .rectangle      import Rectangle, unit
from .monitor        import monitors
from .window_movers  import window_movers


#===========================================================================

class Window(object):
    """
        The Window class is an interface to the Win32 window control
        and placement.

    """

    #-----------------------------------------------------------------------
    # Class attributes to retrieve existing Window objects.

    _windows_by_name = {}
    _windows_by_handle = {}

    #-----------------------------------------------------------------------
    # Class methods to create new Window objects.

    @classmethod
    def get_foreground(cls):
        handle = win32gui.GetForegroundWindow()
        if handle in cls._windows_by_handle:
            return cls._windows_by_handle[handle]
        window = Window(handle=handle)
        return window

    @classmethod
    def get_all_windows(cls):
        def function(handle, argument):
            argument.append(Window(handle))
        argument = []
        win32gui.EnumWindows(function, argument)
        return argument

#    @classmethod
#    def get_window_by_executable(cls, executable):
#        def function(handle, argument):
#            title = windll.user32.GetWindowText(handle)
#            print "title: %r" % title
#        windll.user32.EnumWindows(function, argument)


    #=======================================================================
    # Methods for initialization and introspection.

    def __init__(self, handle):
        self._handle = None
        self._names = set()
        self.handle = handle

    def __str__(self):
        args = ["handle=%d" % self._handle] + list(self._names)
        return "%s(%s)" % (self.__class__.__name__, ", ".join(args))

    #-----------------------------------------------------------------------
    # Methods that control attribute access.

    def _set_handle(self, handle):
        if not isinstance(handle, integer_types):
            raise TypeError("Window handle must be integer or long,"
                            " but received {0!r}".format(handle))
        self._handle = handle
        self._windows_by_handle[handle] = self
    handle = property(fget=lambda self: self._handle,
                      fset=_set_handle,
                      doc="Protected access to handle attribute.")

    def _get_name(self):
        if not self._names:
            return None
        for name in self._names:
            return name
    def _set_name(self, name):
        assert isinstance(name, string_types)
        self._names.add(name)
        self._windows_by_name[name] = self
    name = property(fget=_get_name,
                    fset=_set_name,
                    doc="Protected access to name attribute.")

    #-----------------------------------------------------------------------
    # Direct access to various Win32 methods.

    def _win32gui_func(name):
        func = getattr(win32gui, name)
        return lambda self: func(self._handle)

    _get_rect           = _win32gui_func("GetWindowRect")
    _destroy            = _win32gui_func("DestroyWindow")
    _set_foreground     = _win32gui_func("SetForegroundWindow")
    _bring_to_top       = _win32gui_func("BringWindowToTop")
    _get_window_text    = _win32gui_func("GetWindowText")
    _get_class_name     = _win32gui_func("GetClassName")

    title               = property(fget=_get_window_text)
    classname           = property(fget=_get_class_name)


    def _win32gui_test(name):
        test = getattr(win32gui, name)
        fget = lambda self: test(self._handle) and True or False
        return property(fget=fget,
                        doc="Shortcut to win32gui.%s() function." % name)

    is_valid        = _win32gui_test("IsWindow")
    is_enabled      = _win32gui_test("IsWindowEnabled")
    is_visible      = _win32gui_test("IsWindowVisible")
    is_minimized    = _win32gui_test("IsIconic")
#   is_maximized    = _win32gui_test("IsZoomed")


    def _win32gui_show_window(state):
        return lambda self: win32gui.ShowWindow(self._handle, state)

    minimize        = _win32gui_show_window(win32con.SW_MINIMIZE)
    maximize        = _win32gui_show_window(win32con.SW_MAXIMIZE)
    restore         = _win32gui_show_window(win32con.SW_RESTORE)


    def _get_window_module(self):
        # Get this window's process ID.
        pid = c_ulong()
        windll.user32.GetWindowThreadProcessId(self._handle, pointer(pid))

        # Get the process handle of this window's process ID.
        #  Access permission flags:
        #  0x0410 = PROCESS_QUERY_INFORMATION | PROCESS_VM_READ
        handle = windll.kernel32.OpenProcess(0x0410, 0, pid)

        # Retrieve and return the process's executable path.
        try:
            # Try to use the QueryForProcessImageNameW function
            #  available since Windows Vista.
            buffer_len = c_ulong(256)
            buffer = (c_wchar * buffer_len.value)()
            windll.kernel32.QueryFullProcessImageNameW(handle, 0,
                                                       pointer(buffer),
                                                       pointer(buffer_len))
            buffer = buffer[:]
            buffer = buffer[:buffer.index("\0")]
        except Exception:
            # If the function above failed, fall back to the older
            #  GetModuleFileNameEx function, available since windows XP.
            #  Note that this fallback function seems to fail when
            #  this process is 32 bit Python and handle refers to a
            #  64-bit process.
            buffer_len = 256
            buffer = (c_wchar * buffer_len)()
            windll.psapi.GetModuleFileNameExW(handle, 0, pointer(buffer),
                                              buffer_len)
            buffer = buffer[:]
            buffer = buffer[:buffer.index("\0")]
        finally:
            windll.kernel32.CloseHandle(handle)

        return str(buffer)

    executable = property(fget=_get_window_module)


    #-----------------------------------------------------------------------
    # Methods related to window geometry.

    def get_position(self):
        l, t, r, b = self._get_rect()
        w = r - l; h = b - t
        return Rectangle(l, t, w, h)

    def set_position(self, rectangle):
        assert isinstance(rectangle, Rectangle)
        l, t, w, h = rectangle.ltwh
        win32gui.MoveWindow(self._handle, l, t, w, h, 1)

    def get_containing_monitor(self):
        center = self.get_position().center
        for monitor in monitors:
            if monitor.rectangle.contains(center):
                return monitor
        # Fall through, return first monitor.
        return monitors[0]

    def get_normalized_position(self):
        monitor = self.get_containing_monitor()
        rectangle = self.get_position()
        rectangle.renormalize(monitor.rectangle, unit)
        return rectangle

    def set_normalized_position(self, rectangle, monitor=None):
        if not monitor: monitor = self.get_containing_monitor()
        rectangle.renormalize(unit, monitor.rectangle)
        self.set_position(rectangle)

    #-----------------------------------------------------------------------
    # Methods for miscellaneous window control.

    def set_foreground(self):
        if self.is_minimized:
            self.restore()
        self._set_foreground()

    def move(self, rectangle, animate=None):
        if not animate:
            self.set_position(rectangle)
        else:
            try:
                window_mover = window_movers[animate]
            except KeyError:
                # If the given window mover name isn't found, don't animate.
                self.set_position(rectangle)
            else:
                window_mover.move_window(self, self.get_position(), rectangle)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it 
#   under the terms of the GNU Lesser General Public License as published 
#   by the Free Software Foundation, either version 3 of the License, or 
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but 
#   WITHOUT ANY WARRANTY; without even the implied warranty of 
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public 
#   License along with Dragonfly.  If not, see 
#   <http://www.gnu.org/licenses/>.
#

"""
This file implements the compiler for CMU Pocket Sphinx speech recognition
engine.
"""
from .dragonfly2jsgf import Translator, TranslationState
from ..base import CompilerBase

# noinspection PyUnusedLocal


class SphinxJSGFCompiler(CompilerBase):
    translator = Translator()

    # -----------------------------------------------------------------------
    # Methods for compiling grammars.

    def compile_grammar(self, grammar, *args, **kwargs):
        self._log.debug("%s: Compiling grammar %s." % (self, grammar.name))

        jsgf_grammar = self.translator.translate_grammar(grammar)
        return jsgf_grammar

    def _compile_any_element(self, element, *args, **kwargs):
        state = TranslationState(element)
        expansion = self.translator.get_jsgf_equiv(state).expansion
        return expansion.compile()

    _compile_sequence     = _compile_any_element
    _compile_alternative  = _compile_any_element
    _compile_optional     = _compile_any_element
    _compile_literal      = _compile_any_element
    _compile_rule_ref     = _compile_any_element
    _compile_list_ref     = _compile_any_element
    _compile_dictation    = _compile_any_element
    _compile_impossible   = _compile_any_element
    _compile_empty        = _compile_any_element

    def _compile_unknown_element(self, element, *args, **kwargs):
        raise NotImplementedError("Compiler %s not implemented for element "
                                  "type %s." % (self, element))





from collections                import defaultdict
from ...grammar.grammar_base    import Grammar
from ...grammar.rule_base       import Rule
from ...grammar.elements_basic  import (Sequence, Alternative, RuleRef,
                                        Optional)


#===========================================================================

class CommandFamily(Grammar):

    def __init__(self, name):
        Grammar.__init__(self, name)
        self._extras = []
        self._states = []
        self._toplevel_rules = ()
        self._previous_details = None

    #-----------------------------------------------------------------------

    @property
    def states(self):
        return tuple(self._states)

    def add_states(self, *states):
        self._states.extend(states)

    @property
    def extras(self):
        return tuple(self._extras)

    def add_extras(self, *extras):
        self._extras.extend(extras)

    #-----------------------------------------------------------------------

    def _build_transition_web(self):
        # Build a list of all transitions within this family.
        transitions = []
        for state in self.states:
            transitions.extend(state.transitions)

        # Build a mapping of prestate -> poststate -> list of transitions.
        mapping = defaultdict(lambda: defaultdict(list))
        ids = {}
        for transition in transitions:
            pre_key       = id(transition.prestate)
            post_key      = id(transition.poststate)
            ids[pre_key]  = transition.prestate
            ids[post_key] = transition.poststate
            mapping[pre_key][post_key].append(transition)

#        # DEBUGGING output.
#        for pre_key, inner_mapping in mapping.items():
#            print "pre-state: %s" % ids[pre_key]
#            for post_key, transitions in inner_mapping.items():
#                print "  post-state: %s" % ids[post_key]
#                for transition in transitions:
#                    print "    command: %s" % transition.command

        # Build transition-web container.
        web = {}
        for pre_key, inner_mapping in mapping.items():
            pre_state = ids[pre_key]
            post_state_transitions_pairs = []
            for post_key, transitions in inner_mapping.items():
                post_state = ids[post_key]
                pair = (post_state, transitions)
                post_state_transitions_pairs.append(pair)
            web[pre_key] = (pre_state, post_state_transitions_pairs)

        return web

    def _build_transition_rules(self, web):
        transition_rules = {}
        for (pre_state, post_state_transitions_pairs) in web.values():
            for (post_state, transitions) in post_state_transitions_pairs:
                rule = TransitionsRule(transitions, pre_state, post_state)
                key = (id(pre_state), id(post_state))
                transition_rules[key] = rule
        return transition_rules

    def _build_toplevel_rules(self, web, transition_rules, length):
        toplevel_rules = []
        for (pre_state, post_state_transitions_pairs) in web.values():
            if not pre_state.toplevel:
                continue
            tree = self._build_transition_tree(pre_state, web, length)
            element = self._build_tree_element(tree, transition_rules)
            rule = ToplevelRule(element, pre_state)
            toplevel_rules.append(rule)
        return toplevel_rules

    def _build_transition_tree(self, pre_state, web, length):
        if length <= 0:
            return (pre_state, ())
        branches = []
        for (post_state, transitions) in web[id(pre_state)][1]:
            branch = self._build_transition_tree(post_state, web, length-1)
            branches.append(branch)
        return (pre_state, tuple(branches))

    def _build_tree_element(self, tree, transition_rules):
        pre_state, branches = tree
        if not branches:
            return None
        alternatives = []
        for branch in branches:
            post_state = branch[0]
            key = (id(pre_state), id(post_state))
            transition_rule = transition_rules[key]
            post_element = self._build_tree_element(branch, transition_rules)
            if post_element:
                branch_element = Sequence((RuleRef(transition_rule),
                                           Optional(post_element)))
            else:
                branch_element = RuleRef(transition_rule)
            alternatives.append(branch_element)
        if len(alternatives) == 1:   element = alternatives[0]
        elif len(alternatives) > 1:  element = Alternative(alternatives)
        else:                        element = None
        return element

    #-----------------------------------------------------------------------

    def load(self, length=3):
        web = self._build_transition_web()
        transition_rules = self._build_transition_rules(web)
        toplevel_rules = self._build_toplevel_rules(web, transition_rules, length)
        self._toplevel_rules = toplevel_rules
        for rule in toplevel_rules:
            self.add_rule(rule)

#        for rule in self.rules:
#            print "loading rule: %r (exported %s)" % (rule.name, rule.exported)
#            print
#            print "    ", rule.gstring()
#            print
#
#        self._log.error("Loading %s" % self)
#        print "loading"
#        import time; time.sleep(2)
        Grammar.load(self)

    def unload(self):
        Grammar.unload(self)
        self._toplevel_rules = ()

    #-----------------------------------------------------------------------

    def process_begin(self, executable, title, handle):
        # Trivial optimization: only check for state-activity changes
        #  if the foreground window has changed in some way.
        details = (executable, title, handle)
        if details == self._previous_details:
            return
        self._previous_details = details

        for rule in self._toplevel_rules:
            if rule.pre_state.is_active_toplevel():
#                print "activating:", rule
                rule.activate()
            else:
#                print "de-activating:", rule
                rule.deactivate()


#===========================================================================

class ToplevelRule(Rule):

    def __init__(self, element, pre_state):
        self._pre_state = pre_state
        name = "_ToplevelRule_%s" % pre_state
        Rule.__init__(self, element=element, name=name, exported=True)

    @property
    def pre_state(self):
        return self._pre_state

    def value(self, node):
        transitions = []
        path = [iter(node.children)]
        while path:
            try:
                head = next(path[-1])
            except StopIteration:
                path.pop()
                continue
            if isinstance(head.actor, TransitionsRule):
#                print "transition words:", head.words()
                transitions.append(head.value())
            else:
                path.append(iter(head.children))
#        print "transitions:", transitions
        return transitions

    def process_recognition(self, node):
#        print "processing recognition:", self
        transitions = self.value(node)
        for transition in transitions:
            transition.execute()


#---------------------------------------------------------------------------

class TransitionsRule(Rule):

    def __init__(self, transitions, pre_state, post_state):
        self._pre_state   = pre_state
        self._post_state  = post_state

        # Build list of alternatives from all transition elements.
        alternatives = [t.command.element for t in transitions]
        if len(alternatives) == 1:  element = alternatives[0]
        else:                       element = Alternative(alternatives)

#        for transition in transitions:
#            print "command element", transition.command.element, id (transition.command.element)

        self._transitions_by_id = dict((id(t.command.element), t)
                                       for t in transitions)

        name = "_TransitionsRule_%s_%s" % (pre_state, post_state)
        Rule.__init__(self, element=element, name=name, exported=False)

    def value(self, node):
        transition_node = node.children[0]
        if id(transition_node.actor) not in self._transitions_by_id:
            transition_node = transition_node.children[0]
#        print "transition node actor", transition_node.actor, id(transition_node.actor)
        transition = self._transitions_by_id[id(transition_node.actor)]
#        print "returning down and transition:", transition_node.words()
        return BoundTransition(transition, transition_node)


#---------------------------------------------------------------------------

class BoundTransition(object):

    def __init__(self, transition, node):
        self._transition = transition
        self._node = node

    def execute(self):
        self._transition.execute(self._node)



"""
CMU Pocket Sphinx engine example using a few dragonfly classes
"""

import time
from dragonfly import *


def pause(n):
    print("Pausing for %d seconds" % n)
    g.engine.pause_recognition()
    time.sleep(n)
    g.engine.resume_recognition()
    print("Resuming")


def disconnect():
    print("Disconnecting engine")
    g.engine.disconnect()


class TestRule(MappingRule):
    mapping = {
        "disconnect engine": Function(disconnect),
        "hello <dictation>": ActionBase(),
        "more <dictation> stuff and <dictation>": ActionBase(),
        "pause for <n> seconds": Function(pause),
    }

    extras = [
        Dictation("dictation"),
        IntegerRef("n", 1, 20),
    ]

g = Grammar("test", engine=get_engine("sphinx"))
g.add_rule(TestRule())
g.load()


if __name__ == '__main__':
    # If this file is being run, not imported, set up the engine and an observer
    # to print some info.

    class Observer(RecognitionObserver):
        def on_begin(self):
            print("Speech started.")

        @staticmethod
        def _get_words(words_list):
            # Get just the words from the tuple list
            return " ".join([word for word, _ in words_list])

        def on_recognition(self, words_list):
            print(self._get_words(words_list))

        def on_failure(self):
            print("Sorry, what was that?")

        def on_next_rule_part(self, words_list):
            print("Current words: %s" % self._get_words(words_list))
            print("Awaiting next rule part...")

    observer = Observer()
    observer.register()
    g.engine.recognise_forever()



import unittest

import jsgf.ext

from dragonfly.engines.backend_sphinx.dragonfly2jsgf import *
from dragonfly import *
from dragonfly.parser import ParserError
from dragonfly import List as DragonflyList, DictList as DragonflyDictList


# Some JSGF class aliases for readability
JRule, JSeq, JOpt = jsgf.Rule, jsgf.Sequence, jsgf.OptionalGrouping
JAlt, JRef, JRep, = jsgf.AlternativeSet, jsgf.RuleRef, PatchedRepeat
JDict = jsgf.ext.Dictation


class TranslatorCase(unittest.TestCase):
    translator = Translator()


class PlainCase(TranslatorCase):
    def test_plain(self):
        spec = "Hello world"
        expected = "public <test> = hello world;"
        rule = self.translator.translate_to_rule("test", spec, True)
        self.assertEqual(expected, rule.compile())

    def test_plain_match(self):
        spec = "Hello world"
        rule = self.translator.translate_to_rule("test", spec, True)
        self.assertTrue(rule.matches("hello world"))


class BadRuleCase(TranslatorCase):
    def test_undefined_rule(self):
        spec = "Hello <person>"
        self.assertRaises(Exception, Compound, spec)

    def test_bad_dragonfly_spec(self):
        def is_bad(spec):
            self.assertRaises(ParserError, self.translator.translate, spec)

        is_bad("Hello world[")
        is_bad("Hello world(")
        is_bad("Hello world)")
        is_bad("Hello world#")
        is_bad("Hello world<")
        is_bad("<[a]>")
        is_bad("<a>>")
        is_bad("[]")


class OptionalCase(TranslatorCase):
    def test_simple_optional(self):
        spec = "Hello [glorious] world"
        expected = "public <test> = hello [glorious] world;"
        self.assertEqual(expected,
                         self.translator.translate_to_rule("test", spec, True)
                         .compile())

    def test_complex_optional(self):
        spec = "[[hello] world] hi"
        expected = "public <test> = [[hello] world] hi;"

        self.assertEqual(expected,
                         self.translator.translate_to_rule("test", spec, True)
                         .compile())


class RepetitionCase(TranslatorCase):
    def test_repetition_with_no_max(self):
        r = Rule("test", Repetition(Literal("hello")), exported=True)
        expected = LinkedRule("test", True, JRep("hello"), r)
        actual = self.translator.translate_rule(r).jsgf_rule
        self.assertEqual(actual, expected)

    def test_repetition_with_max(self):
        """Repetition limits are ignored for all rules"""
        r = Rule("test", Repetition(Literal("hello"), max=3), exported=True)
        expected_expansion = JRep("hello")
        expected = LinkedRule("test", True, expected_expansion, r)
        actual = self.translator.translate_rule(r).jsgf_rule
        self.assertEqual(actual, expected)

        # Test again with a higher max value
        r = Rule("test", Repetition(Literal("hello"), max=6), exported=True)
        expected_expansion = JRep("hello")
        expected = LinkedRule("test", True, expected_expansion, r)
        actual = self.translator.translate_rule(r).jsgf_rule
        self.assertEqual(actual, expected)

    def test_with_dictation(self):
        """
        Test that rules involving dictation translate the same regardless of the
        min/max values of the Repetition element.
        """
        r = Rule("test", Repetition(Dictation()), exported=True)
        expected = LinkedRule("test", True, JRep(JDict()), r)
        self.assertEqual(self.translator.translate_rule(r).jsgf_rule, expected)

        r = Rule("test", Repetition(Dictation(), 1, 16), exported=True)
        self.assertEqual(self.translator.translate_rule(r).jsgf_rule, expected)


class ListsCase(TranslatorCase):
    def setUp(self):
        self.fruit_list = DragonflyList("fruit")
        self.fruit_list.append("apple")
        self.fruit_list_ref = ListRef("fruit_ref", self.fruit_list)

        self.dict_list = DragonflyDictList("fruit_dict")
        self.dict_list["apple"] = "yep"
        self.dict_list["mango"] = "nope"
        self.dict_list_ref = DictListRef("fruit_dict_ref", self.dict_list)

    def test_list(self):
        actual = self.translator.translate_list(self.fruit_list)
        expected = JRule("fruit", False, JAlt("apple"))
        self.assertEqual(expected, actual)

    def test_list_ref(self):
        element = self.fruit_list_ref
        state = self.translator.translate_list_ref(TranslationState(element))
        jsgf_rule = JRule("fruit", False, JAlt(*self.fruit_list))
        self.assertEqual(JRef(jsgf_rule), state.expansion)
        self.assertListEqual([jsgf_rule], state.dependencies)

    def test_list_referencing_rule(self):
        element = self.fruit_list_ref
        r = Rule("fav_fruit", element, exported=True)
        state = self.translator.translate_rule(r)
        list_rule = JRule("fruit", False, JAlt(*self.fruit_list))
        expected_rule = LinkedRule("fav_fruit", True, JRef(list_rule), r)
        self.assertEqual(state.jsgf_rule, expected_rule)

    def test_dict_list(self):
        actual = self.translator.translate_dict_list(self.dict_list)
        expected = JRule("fruit_dict", False, JAlt("apple", "mango"))
        self.assertEqual(expected, actual)

    def test_dict_list_ref(self):
        element = self.dict_list_ref
        state = self.translator.translate_dict_list_ref(TranslationState(element))
        expected_rule = JRule("fruit_dict", False, JAlt("apple", "mango"))
        self.assertEqual(JRef(expected_rule), state.expansion)
        self.assertListEqual([expected_rule], state.dependencies)

    def test_dict_list_referencing_rule(self):
        element = self.dict_list_ref
        r = Rule("fruits", element, exported=True)
        state = self.translator.translate_rule(r)
        list_rule = JRule("fruit_dict", False, JAlt("apple", "mango"))
        expected_rule = LinkedRule("fruits", True, JRef(list_rule), r)
        self.assertEqual(state.jsgf_rule, expected_rule)


class ChoiceCase(TranslatorCase):
    """
    Test case for testing whether dragonfly choices compile correctly into
    JSGF
    """
    def test_with_choice(self):
        spec = "Hello <person>"
        compound_rule = CompoundRule(name="greet", spec=spec, extras=[
            Choice("person", {"Bob": "Bob", "John": "John"})
        ])
        expected = "public <greet> = hello (bob|john);"
        rule = self.translator.translate_rule(compound_rule).jsgf_rule
        self.assertEqual(expected, rule.compile())

    def test_choice_match(self):
        spec = "Hello <person>"
        compound_rule = CompoundRule(name="greet", spec=spec, extras=[
            Choice("person", {"Bob": "Bob", "John": "John"})
        ])
        rule = self.translator.translate_rule(compound_rule).jsgf_rule
        self.assertTrue(rule.matches("hello Bob"))


class MappingRuleCase(TranslatorCase):
    def test_int_ref(self):
        r1 = MappingRule(
            name="test",
            mapping={
                "back <n>": Key("backspace:%(n)d"),
                "kill [<n>]": Key("delete:%(n)d"),
            },
            extras=[
                IntegerRef("n", 1, 5)
            ],
            defaults={
                "n": 1
            }
        )

        state = self.translator.translate_rule(r1)
        numbers = ["one", "two", "three", "four"]
        n = JRule("n", False, JAlt(*numbers))
        self.assertListEqual(state.dependencies, [n])
        self.assertEqual(
            state.jsgf_rule,
            JRule("test", True, JAlt(
                JSeq("back", JRef(n)),
                JSeq("kill", JOpt(JRef(n))))))

        # Test that matching works correctly
        self.assertTrue(state.jsgf_rule.matches("kill"))
        for s in numbers:
            self.assertTrue(state.dependencies[0].matches(s))
            self.assertTrue(state.jsgf_rule.matches("back %s" % s),
                            "'back %s' did not match" % s)
            self.assertTrue(state.jsgf_rule.matches("kill %s" % s),
                            "'kill %s' did not match" % s)


class GrammarCase(TranslatorCase):
    def test_simple_rule_grammar(self):
        grammar = Grammar("test")
        rule = CompoundRule(name="test", spec="hello world")
        grammar.add_rule(rule)
        jsgf_grammar = self.translator.translate_grammar(grammar)
        self.assertTrue(jsgf_grammar.rules[0].matches("HELLO WORLD"))

    def test_using_rule_ref(self):
        rule1 = Rule("rule", Literal("hello"), exported=True)
        rule_ref = RuleRef(rule1, name="rule_ref")
        state = TranslationState(rule_ref)
        self.translator.translate_rule_ref(state)
        self.assertEqual(state.element, rule_ref)
        expected_rule = JRule("rule_ref", True, "hello")
        self.assertListEqual(state.dependencies, [expected_rule])
        self.assertEqual(state.expansion, JRef(expected_rule))

    def test_repeated_referenced_dictation(self):
        """
        Test that dragonfly rules containing Dictation referenced by a RuleRef with
        a Repetition element as an ancestor is translated correctly, with rules
        joined as necessary.
        """
        g = Grammar("test")

        # Add two simple rules, a more complex MappingRule and a rule referencing it
        # that allows repetition of its mappings
        r1 = Rule("dict", Dictation("dictation"))
        r2 = Rule("test1", Repetition(RuleRef(r1)), exported=True)
        r3 = MappingRule("mapping", mapping={
            "testing": ActionBase(),
            "hello <dictation>": ActionBase(),
            "<dictation>": ActionBase(),
        }, extras=[Dictation("dictation")])
        r4 = Rule("test2", Repetition(RuleRef(r3)), exported=True)
        g.add_rule(r2)
        g.add_rule(r4)

        # Translate the 'test' grammar and make some assertions
        translated = self.translator.translate_grammar(g)
        expected_test1 = LinkedRule("test1", True, JRep(JDict()), r1)

        # Note: this is not how the engine sees MappingRules with Dictation
        # elements: they are further processed later on into JSGF SequenceRules.
        # This is the format the rules should be in for the later processing.

        expected_test2 = LinkedRule("test2", True, JRep(
            JAlt("testing", JSeq("hello", JDict()), JDict())
        ), r4)

        # Neither of the referenced rules should be in the final grammar, only the
        # two joined rules
        self.assertListEqual(translated.rules, [expected_test1, expected_test2])


if __name__ == '__main__':
    unittest.main()



"""
Tests for the CMU Pocket Sphinx engine

Most engine functionality is tested here, although the tests are done entirely
via `mimic`, so there are some things which have to be tested manually for the
moment.

These tests assume the US English pronunciation dictionary, acoustic and language
models distributed with the `pocketsphinx` Python package are used.
"""

import unittest

import time
import logging

from sphinxwrapper import DefaultConfig

from dragonfly import *
from dragonfly import List as DragonflyList, DictList as DragonflyDict
from dragonfly.engines.backend_sphinx.engine import SphinxEngine
from dragonfly.test import TestContext


class MockLoggingHandler(logging.Handler):
    """
    Mock logging handler to check for expected logs.
    Adapted this from a Stack Overflow answer: https://stackoverflow.com/a/1049375
    """

    def __init__(self, *args, **kwargs):
        self.messages = {
            'debug': [],
            'info': [],
            'warning': [],
            'error': [],
            'critical': [],
        }
        logging.Handler.__init__(self, *args, **kwargs)

    def emit(self, record):
        self.messages[record.levelname.lower()].append(record.getMessage())


class SphinxEngineCase(unittest.TestCase):
    """
    Base TestCase class for Sphinx engine tests
    """
    def setUp(self):
        engine = get_engine("sphinx")
        assert isinstance(engine, SphinxEngine)
        assert engine.name == "sphinx"
        self.engine = engine

        # Save current config
        self.engine_config = self.engine.config

        # Set a default NEXT_PART_TIMEOUT value so there aren't weird timing errors
        # that break tests. Value of 0 means no timeout at all.
        self.engine.config.NEXT_PART_TIMEOUT = 0

        # Set training data directory to None for the tests.
        self.engine.config.TRAINING_DATA_DIR = None

        # Ensure the relevant default configuration values are used
        self.engine.config.START_ASLEEP = False
        self.engine.config.WAKE_PHRASE = "wake up"
        self.engine.config.SLEEP_PHRASE = "go to sleep"
        self.engine.config.START_TRAINING_PHRASE = "start training session"
        self.engine.config.END_TRAINING_PHRASE = "end training session"
        self.engine.config.LANGUAGE = "en"

        # Map for test functions
        self.test_map = {}

        # Add a logging handler.
        self.logging_handler = MockLoggingHandler()
        self.engine.log.addHandler(self.logging_handler)

        # Connect the engine.
        engine.connect()

    def tearDown(self):
        # Restore saved config and do some other things.
        self.engine.config = self.engine_config
        self.engine.disconnect()
        self.test_map.clear()
        self.engine.log.removeHandler(self.logging_handler)

    # ---------------------------------------------------------------------
    # Methods for control-flow assertion.

    def get_test_function(self):
        """
        Create and return a test function used for testing whether rules are
        processed correctly, insofar as they reach the correct processing method.

        Note that returned test functions accept variable arguments.
        :return: callable
        """
        def func(*args):
            # Function was reached
            try:
                self.test_map[id(func)] += 1
            except KeyError:
                # Ignore any key errors; if this function's id is not in test_map,
                # then it's not relevant to the currently running test method.
                pass

        self.test_map[id(func)] = 0
        return func

    def assert_test_function_called(self, func, n):
        """
        Assert that a test function was called n times.
        :type func: callable
        :param n: number of times the test function should have been called
        """
        x = self.test_map[id(func)]
        self.assertEqual(x, n, "wrapped test function was called %d time(s) "
                               "instead of %d time(s)" % (x, n))

    def reset_test_functions(self):
        """
        Reset the test_map values for all test functions.
        """
        for key in self.test_map.keys():
            self.test_map[key] = 0

    # ---------------------------------------------------------------------
    # Methods for asserting mimic success or failure

    def assert_mimic_success(self, *phrases):
        """
        Assert that the engine can successfully mimic a number of speech strings,
        or fail with the provided error message.
        """
        try:
            self.engine.mimic_phrases(*phrases)
        except MimicFailure:
            self.fail("MimicFailure caught")

    def assert_mimic_failure(self, *phrases):
        self.assertRaises(MimicFailure, self.engine.mimic_phrases, *phrases)


class BasicEngineTests(SphinxEngineCase):
    """
    Tests for basic engine functionality.
    """
    def test_get_engine_sphinx_is_usable(self):
        """
        Verify that the sphinx engine is usable by testing that a simple
        rule is loaded correctly and works correctly.
        """
        test = self.get_test_function()

        class TestRule(CompoundRule):
            spec = "hello world"
            _process_recognition = test

        g = Grammar("test1")
        g.add_rule(TestRule())
        g.load()
        self.assertTrue(g.loaded)
        self.assert_mimic_success("hello world")
        self.assert_test_function_called(test, 1)

    def test_engine_config(self):
        """
        Test whether engine configuration is validated correctly.
        """

        # Required config names
        required = [
            "DECODER_CONFIG",
            "LANGUAGE",
            "PYAUDIO_STREAM_KEYWORD_ARGS",
            "NEXT_PART_TIMEOUT",
            "START_ASLEEP",
            "WAKE_PHRASE",
            "SLEEP_PHRASE",
            "WAKE_PHRASE_THRESHOLD",
            "SLEEP_PHRASE_THRESHOLD",
            "TRAINING_DATA_DIR",
            "START_TRAINING_PHRASE",
            "START_TRAINING_THRESHOLD",
            "END_TRAINING_PHRASE",
            "END_TRAINING_THRESHOLD",
        ]

        class TestConfig(object):
            DECODER_CONFIG = DefaultConfig()
            LANGUAGE = "en"
            PYAUDIO_STREAM_KEYWORD_ARGS = {}
            NEXT_PART_TIMEOUT = 0
            START_ASLEEP = False
            WAKE_PHRASE = "wake up"
            WAKE_PHRASE_THRESHOLD = 1e-20
            SLEEP_PHRASE = "go to sleep"
            SLEEP_PHRASE_THRESHOLD = 1e-40
            TRAINING_DATA_DIR = None
            START_TRAINING_PHRASE = "start training session"
            START_TRAINING_THRESHOLD = 1e-48
            END_TRAINING_PHRASE = "end training session"
            END_TRAINING_THRESHOLD = 1e-45

        def set_config(value):
            self.engine.config = value

        # Test with correct attributes
        self.assertIsNone(set_config(TestConfig))

        # Delete each attribute and check that an EngineError is raised upon
        # setting the config again.
        for name in required:
            delattr(TestConfig, name)
            self.assertRaises(EngineError, set_config, TestConfig)

    def test_load_unload(self):
        """
        Test that grammars are properly loaded and unloaded.
        """
        grammar = Grammar("test")
        # Add the simplest rule and test that it loads and unloads correctly
        rule1 = CompoundRule("rule1", "hello")
        grammar.add_rule(rule1)
        grammar.load()
        self.assertTrue(grammar.loaded)
        self.assert_mimic_success("hello")
        grammar.unload()
        self.assertFalse(grammar.loaded)
        self.assert_mimic_failure("hello")

        # Test that a dependent rule is unloaded correctly.
        rule2 = CompoundRule("rule2", "<rule1ref> there",
                             extras=[RuleRef(rule1, name="rule1ref")])
        grammar.add_rule(rule2)
        grammar.load()
        self.assertTrue(grammar.loaded)
        self.assert_mimic_success("hello there")
        grammar.unload()
        self.assertFalse(grammar.loaded)
        self.assert_mimic_failure("hello there")

    def test_mapping_rule(self):
        """
        Test that the engine can handle a grammar with a MappingRule.
        """
        # Get two test functions for the rules to use
        test1 = self.get_test_function()
        test2 = self.get_test_function()

        class TestRule(MappingRule):
            mapping = {
                "hello": Function(test1),
                "hi": Function(test2),
            }

        g = Grammar("test")
        g.add_rule(TestRule())
        g.load()
        self.assertTrue(g.loaded)

        # Test 'hello' rule works and that only the test1 function was called
        self.assert_mimic_success("hello")
        self.assert_test_function_called(test1, 1)
        self.assert_test_function_called(test2, 0)
        # Reset test functions
        self.reset_test_functions()

        # Test 'hi' rule works and that only the test2 function was called
        self.assert_mimic_success("hi")
        self.assert_test_function_called(test1, 0)
        self.assert_test_function_called(test2, 1)

    def test_update_list(self):
        """
        Test support for dragonfly lists.
        """
        # Set up a new list and add an item
        fruit_list = DragonflyList("fruit")
        fruit_list.append("apple")
        fruit_list_ref = ListRef("fruit_ref", fruit_list)

        # Load a new grammar with a rule referencing the fruit list
        grammar = Grammar("test")
        r = Rule("fav_fruit", fruit_list_ref, exported=True)
        grammar.add_rule(r)
        grammar.load()
        self.assertTrue(grammar.loaded)

        self.assert_mimic_success("apple")

        # Try updating the list
        fruit_list.append("banana")
        self.assertListEqual(["apple", "banana"], r.element.list)

        # If mimic fails, then the list wasn't updated correctly.
        self.assert_mimic_success("banana")

    def test_update_list_dict(self):
        """
        Test support for dragonfly dict lists.
        """
        # Set up a new dict list and add an item
        fruit_dict = DragonflyDict("fruit_dict")
        fruit_dict["mango"] = "nope"

        # Add a new rule to the grammar using a DictListRef element and load it
        grammar = Grammar("test")
        r = Rule("fruits", DictListRef("fruit_dict_ref", fruit_dict),
                 exported=True)
        grammar.add_rule(r)
        grammar.load()
        self.assertTrue(grammar.loaded)

        self.assert_mimic_success("mango")

        # Try updating the list
        fruit_dict["mandarin"] = "yep"
        self.assertDictEqual({"mango": "nope", "mandarin": "yep"}, r.element.list)

        # If mimic fails, then the dict list wasn't updated correctly.
        self.assert_mimic_success("mandarin")

    def test_choices(self):
        """
        Test that a dragonfly rule using a Choice extra works correctly.
        """
        test = self.get_test_function()
        digits_dict = {
            "one": "1",
            "two": "2",
            "three": "3",
            "four": "4",
            "five": "5",
            "six": "6",
            "seven": "7",
            "eight": "8",
            "nine": "9",
            "point": ".",
        }

        class TestRule(CompoundRule):
            spec = "<digits>"
            extras = [Choice("digits", digits_dict)]
            _process_recognition = test

        g = Grammar("test")
        g.add_rule(TestRule())
        g.load()
        self.assertTrue(g.loaded)

        # Test that all choices are processed correctly
        for i, phrase in enumerate(["one", "two", "three", "four", "five", "six",
                                    "seven", "eight", "nine", "point"]):
            self.assert_mimic_success(phrase)
            self.assert_test_function_called(test, i+1)  # starting at 1, not 0.

    def test_recognition_history(self):
        observer = RecognitionHistory()
        observer.register()
        self.assertListEqual(observer, [])  # RecognitionHistory is a list subclass

        # Set up a test grammar and rule
        g = Grammar("test")
        g.add_rule(CompoundRule("rule1", "testing"))
        g.load()
        self.assertTrue(g.loaded)

        self.assert_mimic_success("testing")
        self.assertListEqual(observer, [[("testing", 0)]])

    def test_name_conflict(self):
        """
        Test whether the engine correctly handles rules with the same name, but
        different elements.
        """
        g1 = Grammar("test1")
        g2 = Grammar("test2")
        g1.add_rule(CompoundRule(name="rule", spec="test"))
        g2.add_rule(CompoundRule(name="rule", spec="testing"))

        g1.load()
        self.assertTrue(g1.loaded)
        g2.load()
        self.assertTrue(g2.loaded)
        self.assert_mimic_success("test")
        self.assert_mimic_success("testing")

    def test_same_rule_in_different_grammars(self):
        g1 = Grammar("test1")
        g2 = Grammar("test2")

        test1 = self.get_test_function()
        test2 = self.get_test_function()

        class Rule1(CompoundRule):
            spec = "test"
            _process_recognition = test1

        class Rule2(CompoundRule):
            spec = "test"
            _process_recognition = test2

        # Add and load the rules
        g1.add_rule(Rule1())
        g2.add_rule(Rule2())
        g1.load()
        self.assertTrue(g1.loaded)
        g2.load()
        self.assertTrue(g2.loaded)

        self.assert_mimic_success("test")

        # Check that only one of these rules was processed
        self.assert_test_function_called(test1, 1)
        self.assert_test_function_called(test2, 0)

    def test_pause_resume_recognition(self):
        test = self.get_test_function()

        class TestRule(CompoundRule):
            spec = "hello world"
            _process_recognition = test

        g = Grammar("test1")
        g.add_rule(TestRule())
        g.load()
        self.assertTrue(g.loaded)
        self.assertFalse(self.engine.recognition_paused)

        # Mimicking hello world should fail *silently* when recognition is paused
        self.engine.pause_recognition()
        self.assertTrue(self.engine.recognition_paused)
        self.engine.mimic("hello world")  # note this is *not* a recognition failure
        self.assert_test_function_called(test, 0)
        self.engine.resume_recognition()
        self.assertFalse(self.engine.recognition_paused)
        self.assert_mimic_success("hello world")
        self.assert_test_function_called(test, 1)

        # Test again mimicking wake and sleep phrases
        self.assert_mimic_success("go to sleep")
        self.assertTrue(self.engine.recognition_paused)
        self.engine.mimic("hello world")
        self.assert_test_function_called(test, 1)  # no change
        self.assert_mimic_success("wake up")
        self.assertFalse(self.engine.recognition_paused)
        self.assert_mimic_success("hello world")
        self.assert_test_function_called(test, 2)

    def test_rule_contexts(self):
        # Test that rules with contexts work and that rules with context=None still
        # work regardless of context.
        test1, test2 = self.get_test_function(), self.get_test_function()
        rule1 = MappingRule(name="test1", mapping={"test global": Function(test1)})
        context = TestContext(True)
        rule2 = MappingRule(
            name="test2", mapping={"test context": Function(test2)},
            context=TestContext(True)
        )

        # Create and load a grammar.
        grammar = Grammar("test")
        grammar.add_rule(rule1)
        grammar.add_rule(rule2)
        grammar.load()

        # Test that both rules rules work.
        self.assert_mimic_success("test context")
        self.assert_mimic_success("test global")
        self.assert_test_function_called(test1, 1)
        self.assert_test_function_called(test2, 1)

        # Go out of context and test both rules again.
        context.active = False
        self.assert_mimic_failure("test context")
        self.assert_mimic_success("test global")
        self.assert_test_function_called(test1, 2)
        self.assert_test_function_called(test2, 1)

        # Try again in-context.
        context.active = True
        self.assert_mimic_success("test context")
        self.assert_mimic_success("test global")
        self.assert_test_function_called(test1, 3)
        self.assert_test_function_called(test2, 2)

    def test_grammar_contexts(self):
        # Test that grammars with contexts work and that global contexts still work
        # regardless of context.
        test1, test2 = self.get_test_function(), self.get_test_function()

        class TestRule1(MappingRule):
            mapping = {"test global": Function(test1)}

        class TestRule2(MappingRule):
            mapping = {"test context": Function(test2)}

        # Create and load a global grammar.
        grammar1 = Grammar("global")
        grammar1.add_rule(TestRule1())
        grammar1.load()

        # Create and load a grammar using a context.
        context = TestContext(active=True)
        grammar2 = Grammar("context", context=context)
        grammar2.add_rule(TestRule2())
        grammar2.load()

        # Test that rules in both grammars work.
        self.assert_mimic_success("test context")
        self.assert_mimic_success("test global")
        self.assert_test_function_called(test1, 1)
        self.assert_test_function_called(test2, 1)

        # Go out of context and test both rules again.
        context.active = False
        self.assert_mimic_failure("test context")
        self.assert_mimic_success("test global")
        self.assert_test_function_called(test1, 2)
        self.assert_test_function_called(test2, 1)

        # Try again in-context.
        context.active = True
        self.assert_mimic_success("test context")
        self.assert_mimic_success("test global")
        self.assert_test_function_called(test1, 3)
        self.assert_test_function_called(test2, 2)

    def test_start_asleep(self):
        # config.START_ASLEEP is False for the tests by default, so test that first.
        self.assertFalse(self.engine.recognition_paused)
        self.assert_mimic_success("go to sleep")

        # Now set it to True, restart the engine and test again.
        self.engine.config.START_ASLEEP = True
        self.engine.disconnect()
        self.engine.connect()
        self.engine.post_loader_init()  # used instead of 'recognise_forever' here
        self.assertTrue(self.engine.recognition_paused)
        self.assert_mimic_success("wake up")

    def test_key_phrases(self):
        """
        Test key phrase functionality and observer notify methods.
        """
        test1 = self.get_test_function()
        test2 = self.get_test_function()
        on_begin_test = self.get_test_function()
        on_recognition_test = self.get_test_function()
        on_failure_test = self.get_test_function()

        # Set up a custom observer using test methods
        class TestObserver(RecognitionObserver):
            on_begin = on_begin_test
            on_recognition = on_recognition_test
            on_failure = on_failure_test

        self.engine.register_recognition_observer(TestObserver())

        # Register two key phrases
        self.engine.set_keyphrase("hello world", 1e-20, test1)
        self.engine.set_keyphrase("testing testing", 1e-30, test2)

        # Test that both key phrases can be mimicked
        self.assert_mimic_success("hello world")
        self.assert_test_function_called(test1, 1)
        self.assert_test_function_called(on_begin_test, 1)
        self.assert_test_function_called(on_recognition_test, 1)
        self.assert_mimic_success("testing testing")
        self.assert_test_function_called(test2, 1)
        self.assert_test_function_called(on_begin_test, 2)
        self.assert_test_function_called(on_recognition_test, 2)
        self.assert_test_function_called(on_failure_test, 0)

        # Test that neither of the key phrases match when recognition is paused
        self.engine.pause_recognition()
        self.engine.mimic("hello world")
        self.assert_test_function_called(test1, 1)  # no change
        self.engine.mimic("testing testing")
        self.assert_test_function_called(test2, 1)  # no change

        # Test that no notify test functions have been called again
        self.assert_test_function_called(on_begin_test, 2)
        self.assert_test_function_called(on_recognition_test, 2)
        self.assert_test_function_called(on_failure_test, 0)

        # Resume recognition again. on_recognition_test should have been called.
        self.engine.resume_recognition()
        self.assert_test_function_called(on_recognition_test, 3)

        # Test that removed key phrases no longer match
        self.engine.unset_keyphrase("hello world")
        self.assert_mimic_failure("hello world")
        self.assert_test_function_called(test1, 1)  # no change
        self.assert_test_function_called(on_begin_test, 3)
        self.assert_test_function_called(on_failure_test, 1)
        self.engine.unset_keyphrase("testing testing")
        self.assert_mimic_failure("testing testing")
        self.assert_test_function_called(test2, 1)  # no change
        self.assert_test_function_called(on_begin_test, 4)
        self.assert_test_function_called(on_failure_test, 2)

    def test_unknown_keyphrase_words(self):
        # Test that keyphrases with unknown words log appropriate error messages.
        self.engine.set_keyphrase("notaword", 1e-20, lambda: None)
        self.assertEqual(
            self.logging_handler.messages["error"][0],
            "keyphrase used words not found in the pronunciation dictionary: "
            "notaword"
        )

        # Test invalid built-in keyphrases.
        self.engine.config.WAKE_PHRASE = "wake up unknownword"
        self.engine.config.SLEEP_PHRASE = "aninvalid sleepphrase"
        self.engine.config.START_TRAINING_PHRASE = "another invalidphrase"
        self.engine.config.END_TRAINING_PHRASE = "end trainingsession"

        # Restart the engine manually to verify that an error is logged for the
        # keyphrases on connect().
        self.engine.disconnect()
        self.engine.connect()

        # Check the logged messages. Each of the unknown words in all four
        # built-in keyphrases should be listed.
        self.assertEqual(
            self.logging_handler.messages["error"][1],
            "keyphrase used words not found in the pronunciation dictionary: "
            "aninvalid, invalidphrase, sleepphrase, trainingsession, unknownword"
        )

    def test_unknown_grammar_words(self):
        # Test that grammars using unknown words log appropriate error messages when
        # they fail to load.
        class TestRule1(CompoundRule):
            spec = "testing unknownword"

        class TestRule2(MappingRule):
            mapping = {
                "wordz": ActionBase(),
                "natlink": ActionBase()
            }

        g = Grammar("test")
        g.add_rule(TestRule1())
        g.add_rule(TestRule2())
        g.load()

        # Check the logged messages. The words should be in alphabetical order.
        self.assertEqual(
            self.logging_handler.messages["error"][0],
            "grammar 'test' used words not found in the pronunciation dictionary: "
            "natlink, unknownword, wordz"
        )

    def test_training_session(self):
        # Test that no recognition processing is done when a training session is
        # active.

        # Set up a rule to "train".
        test = self.get_test_function()

        class TestRule(CompoundRule):
            spec = "test training"
            _process_recognition = test

        # Create and load a grammar with the rule.
        grammar = Grammar("test")
        grammar.add_rule(TestRule())
        grammar.load()

        # Set up a custom observer using test methods
        on_begin_test = self.get_test_function()
        on_recognition_test = self.get_test_function()
        on_failure_test = self.get_test_function()

        class TestObserver(RecognitionObserver):
            on_begin = on_begin_test
            on_recognition = on_recognition_test
            on_failure = on_failure_test

        self.engine.register_recognition_observer(TestObserver())

        # Start a training session.
        self.engine.start_training_session()

        # Test that mimic succeeds, no processing occurs, and the TestObserver
        # is still notified of events.
        self.assert_mimic_success("test training")
        self.assert_test_function_called(test, 0)
        self.assert_test_function_called(on_begin_test, 1)
        self.assert_test_function_called(on_recognition_test, 1)
        self.assert_test_function_called(on_failure_test, 0)

        # End the session and test again.
        self.engine.end_training_session()
        self.assert_mimic_success("test training")
        self.assert_test_function_called(test, 1)
        self.assert_test_function_called(on_begin_test, 2)
        self.assert_test_function_called(on_recognition_test, 2)
        self.assert_test_function_called(on_failure_test, 0)


class DictationEngineTests(SphinxEngineCase):
    """
    Tests for the engine's dictation functionality.
    """
    def test_single_dictation(self):
        """
        Test that the engine can handle a dragonfly rule using a Dictation element.
        """
        test = self.get_test_function()

        class TestRule(MappingRule):
            mapping = {"<dictation>": Function(test)}
            extras = [Dictation("dictation")]

        g = Grammar("test")
        g.add_rule(TestRule())
        g.load()
        self.assertTrue(g.loaded)
        self.assert_mimic_success("hello")
        self.assert_test_function_called(test, 1)

        # Test that it works again
        self.assert_mimic_success("hello")
        self.assert_test_function_called(test, 2)

        # Test again with multiple words
        self.assert_mimic_success("hello world")
        self.assert_test_function_called(test, 3)

    def test_dictation_with_other_elements(self):
        """
        Test that the engine can handle dragonfly rules using dictation with other
        elements.

        This is being tested because the way dictation works for this engine is
        different than how the other engines handle it: Sphinx needs utterance
        pauses between grammar and dictation rule parts.
        """
        test1 = self.get_test_function()
        test2 = self.get_test_function()

        class TestRule(MappingRule):
            mapping = {
                "hello <dictation>": Function(test1),
                "<dictation> testing": Function(test2),
            }
            extras = [Dictation("dictation")]

        g = Grammar("test")
        g.add_rule(TestRule())
        g.load()
        self.assertTrue(g.loaded)

        # Dictation and other elements are processed separately with pauses
        self.assert_mimic_success("hello", "world")  # test mapping 1
        self.assert_test_function_called(test1, 1)
        self.assert_test_function_called(test2, 0)

        # Test with mapping 2
        self.assert_mimic_success("testing", "testing")

        # Test that it works again with mapping 1
        self.reset_test_functions()
        self.assert_mimic_success("hello", "world")
        self.assert_test_function_called(test1, 1)
        self.assert_test_function_called(test2, 0)

        # Test with mapping 2 again
        self.reset_test_functions()
        self.assert_mimic_success("test", "testing")
        self.assert_test_function_called(test1, 0)
        self.assert_test_function_called(test2, 1)

        # Test again with multiple words
        self.reset_test_functions()
        self.assert_mimic_success("hello", "hi there")
        self.assert_test_function_called(test1, 1)
        self.assert_test_function_called(test2, 0)

        # And test again using mapping 2
        self.reset_test_functions()
        self.assert_mimic_success("test test test", "testing")
        self.assert_test_function_called(test2, 1)
        self.assert_test_function_called(test1, 0)

        # Test incomplete sequences
        self.reset_test_functions()
        self.assert_mimic_success("hello")  # start of mapping one
        self.assert_test_function_called(test1, 0)  # incomplete sequence
        self.assert_mimic_success("test", "testing")
        self.assert_test_function_called(test1, 1)  # mapping one completely matched
        # mapping two only partially matched with <dictation> as "testing"
        self.assert_test_function_called(test2, 0)

    def test_recognition_observer(self):
        """
        Test that the engine's recognition observer manager works correctly.
        """
        on_begin_test = self.get_test_function()
        on_recognition_test = self.get_test_function()
        on_failure_test = self.get_test_function()
        on_next_rule_part_test = self.get_test_function()

        # Set up a custom observer using test methods
        class TestObserver(RecognitionObserver):
            on_begin = on_begin_test
            on_next_rule_part = on_next_rule_part_test
            on_recognition = on_recognition_test
            on_failure = on_failure_test

        # Set up a TestObserver instance and a grammar with multiple rules to use
        observer = TestObserver()
        observer.register()
        grammar = Grammar("test")
        grammar.add_rule(CompoundRule("rule1", "hello world"))
        grammar.add_rule(CompoundRule("rule2", "say <dictation>",
                                      extras=[Dictation("dictation")]))
        grammar.load()
        self.assertTrue(grammar.loaded)

        # Test that each method is called properly
        self.assert_mimic_success("hello world")

        # on_begin is called during each mimic. on_recognition should be called
        # once per successful and complete recognition. Both on_failure and
        # on_next_rule_part shouldn't have been called yet.
        self.assert_test_function_called(on_begin_test, 1)
        self.assert_test_function_called(on_recognition_test, 1)
        self.assert_test_function_called(on_failure_test, 0)
        self.assert_test_function_called(on_next_rule_part_test, 0)

        # Test with a dictation rule
        self.assert_mimic_success("say")

        # Recognition begins again, is incomplete and no failure yet.
        self.assert_test_function_called(on_begin_test, 2)
        self.assert_test_function_called(on_recognition_test, 1)
        self.assert_test_function_called(on_failure_test, 0)

        # on_next_rule_part should be called because there are more rule parts
        self.assert_test_function_called(on_next_rule_part_test, 1)

        # Test the next part of the dictation rule
        self.assert_mimic_success("testing testing")

        # Recognition begins again, is complete, and no failure yet.
        self.assert_test_function_called(on_begin_test, 3)
        self.assert_test_function_called(on_recognition_test, 2)
        self.assert_test_function_called(on_failure_test, 0)

        # on_next_rule_part shouldn't have been called because this is the last part
        # and on_recognition will be called instead
        self.assert_test_function_called(on_next_rule_part_test, 1)

        # Recognition begins again and failure occurs.
        self.assert_mimic_failure("testing")
        self.assert_test_function_called(on_begin_test, 4)
        self.assert_test_function_called(on_next_rule_part_test, 1)  # no change
        self.assert_test_function_called(on_failure_test, 1)
        self.assert_test_function_called(on_recognition_test, 2)

        # Test that using None or "" also calls the on_failure method
        self.assert_mimic_failure(None)
        self.assert_test_function_called(on_failure_test, 2)
        self.assert_mimic_failure("")
        self.assert_test_function_called(on_failure_test, 3)
        self.assert_test_function_called(on_next_rule_part_test, 1)  # no change

        # Unregister the observer
        observer.unregister()

    def test_no_hypothesis(self):
        """
        Check that if something that doesn't match any rule is mimicked, nothing
        gets recognised.
        """
        test1 = self.get_test_function()
        test2 = self.get_test_function()

        class TestRule(MappingRule):
            mapping = {
                "testing": Function(test1),
                "<dictation>": Function(test2)
            }
            extras = [Dictation("dictation")]

        g = Grammar("test")
        g.add_rule(TestRule())
        g.load()
        self.assertTrue(g.loaded)

        self.assert_mimic_failure(None)
        self.assert_test_function_called(test1, 0)
        self.assert_test_function_called(test2, 0)
        self.assert_mimic_failure("")
        self.assert_test_function_called(test1, 0)
        self.assert_test_function_called(test2, 0)
        self.assert_mimic_success("hello")
        self.assert_test_function_called(test1, 0)
        self.assert_test_function_called(test2, 1)

    def test_sequence_timeout(self):
        """
        Test that when speaking subsequent parts of rules involving Dictation
        elements, the recognition times out if there is a timeout period set.
        """
        # Set a timeout period for this test of 100 ms, which is unreasonable for
        # humans, but fine for mimic().
        timeout = 0.1
        self.engine.config.NEXT_PART_TIMEOUT = timeout

        # Set up a test grammar with some rules
        test1 = self.get_test_function()
        test2 = self.get_test_function()

        class TestRule(MappingRule):
            mapping = {
                "say <dictation>": Function(test1),
                "hello world": Function(test2),
            }
            extras = [Dictation("dictation")]

        grammar = Grammar("test")
        grammar.add_rule(TestRule())
        grammar.load()
        self.assertTrue(grammar.loaded)

        # Test that mapping 1 can be recognised fully
        self.assert_mimic_success("say", "hello")
        self.assert_test_function_called(test1, 1)

        # Start recognising mapping 1 again, then sleep for a bit
        self.assert_mimic_success("say")
        time.sleep(timeout)

        # The rest of mapping 1 should not match
        self.assert_mimic_failure("testing")
        self.assert_test_function_called(test1, 1)  # still only called 1 time

        # Should be able to match mapping 2 now
        self.assert_mimic_success("hello world")
        self.assert_test_function_called(test2, 1)

        # Test that it works with a break shorter than the timeout value
        self.assert_mimic_success("say")
        time.sleep(timeout / 2)
        self.assert_mimic_success("hello")
        self.assert_test_function_called(test1, 2)

        # Test with a timeout of 0 (no timeout)
        timeout = 0
        self.engine.config.NEXT_PART_TIMEOUT = timeout
        self.assert_mimic_success("say")
        time.sleep(0.1)  # sleep for 100ms
        self.assert_test_function_called(test1, 2)  # no change
        self.assert_mimic_success("testing")
        self.assert_test_function_called(test1, 3)

        # Test without a sleep between rule parts
        self.assert_mimic_success("say", "testing")
        self.assert_test_function_called(test1, 4)

        # Test that mapping 2 still has no issue matching
        self.assert_mimic_success("hello world")
        self.assert_test_function_called(test2, 2)

    def test_long_sequence_rule(self):
        """
        Test that the engine can recognise a long sequence rule and that the
        timeout functionality works correctly.
        """
        timeout = 0.1
        self.engine.config.NEXT_PART_TIMEOUT = timeout

        # Set up a test grammar with some rules
        test1 = self.get_test_function()
        test2 = self.get_test_function()

        class TestRule(MappingRule):
            mapping = {
                "testing": Function(test1),
                "testing <dictation> testing <dictation>": Function(test2),
            }
            extras = [Dictation("dictation")]

        grammar = Grammar("test")
        grammar.add_rule(TestRule())
        grammar.load()
        self.assertTrue(grammar.loaded)

        # Test that mapping 2 can be recognised fully
        self.assert_mimic_success("testing", "hello", "testing")
        self.assert_mimic_success("hello")
        self.assert_test_function_called(test2, 1)
        self.assert_test_function_called(test1, 0)

        # Start recognising mapping 2 again, then sleep for a bit
        self.assert_mimic_success("testing")
        time.sleep(timeout + 0.1)

        # Only mapping 1 should have been processed
        self.assert_test_function_called(test1, 1)
        self.assert_test_function_called(test2, 1)

        # Test that it works with a break shorter than the timeout value
        self.assert_mimic_success("testing")
        time.sleep(timeout / 2)
        self.assert_mimic_success("hello")
        time.sleep(timeout / 2)
        self.assert_mimic_success("testing")
        time.sleep(timeout / 2)
        self.assert_mimic_success("hello")
        self.assert_test_function_called(test2, 2)

        # Test with a timeout of 0 (no timeout)
        timeout = 0
        self.engine.config.NEXT_PART_TIMEOUT = timeout
        self.assert_mimic_success("testing")
        time.sleep(0.1)  # sleep for 100ms
        self.assert_test_function_called(test2, 2)  # no changes

        # Test that mapping 1 won't process any more, even after a time.
        self.assert_test_function_called(test1, 1)

        # Mimicking the rest of the rule parts should make the rule process.
        self.assert_mimic_success("hello", "testing", "hello")
        self.assert_test_function_called(test2, 3)

    def test_engine_mimic(self):
        """The engine.mimic method matches any rule/mapping in one call."""
        # Note: it is normal for action.exec errors to be logged for this test.
        # Set up a test grammar with some rules
        test1 = self.get_test_function()
        test2 = self.get_test_function()
        test3 = self.get_test_function()
        test4 = self.get_test_function()

        class TestRule(MappingRule):
            mapping = {
                "testing": Function(test1),
                "testing <dictation> testing": Function(test2),
                "test with multiple words": Function(test3),
                "forward [<n>]": Function(test4)
            }
            extras = [
                Dictation("dictation"),
                IntegerRef("n", 1, 3)
            ]

        grammar = Grammar("test")
        grammar.add_rule(TestRule())
        grammar.load()

        def exec_mimic_action(words):
            # Pass each word to Mimic
            Mimic(*words.split(" ")).execute()

        # Test with the engine's mimic method and the Mimic action
        for f in [self.engine.mimic, exec_mimic_action]:
            self.reset_test_functions()

            # Test mapping 1
            f("testing")
            # engine.mimic will not match speech in sequence like
            # engine.mimic_phrases
            self.assert_test_function_called(test1, 1)
            self.assert_test_function_called(test2, 0)
            self.assert_test_function_called(test3, 0)

            # Test that partial matches of mapping 2 fail.
            if f is exec_mimic_action:
                # ActionBase.execute() catches and logs raised errors, so don't use
                # assertRaises for Mimic(...).execute()
                f("testing testing")
                f("testing hello")
            else:
                # engine.mimic(..) will raise MimicFailure
                self.assertRaises(MimicFailure, f, "testing testing")
                self.assertRaises(MimicFailure, f, "testing hello")

            # Neither test function should have been called
            self.assert_test_function_called(test1, 1)
            self.assert_test_function_called(test2, 0)
            self.assert_test_function_called(test3, 0)

            # Test that full matches of mapping 2 are successful
            f("testing hello testing")
            self.assert_test_function_called(test1, 1)
            self.assert_test_function_called(test2, 1)
            self.assert_test_function_called(test3, 0)

            # Test mapping 3
            f("test with multiple words")
            self.assert_test_function_called(test1, 1)
            self.assert_test_function_called(test2, 1)
            self.assert_test_function_called(test3, 1)  # mapping 3 processed
            self.assert_test_function_called(test4, 0)

            # Test mapping 4
            f("forward")
            self.assert_test_function_called(test4, 1)
            f("forward one")
            self.assert_test_function_called(test4, 2)


# ---------------------------------------------------------------------


if __name__ == '__main__':
    unittest.main()



# This file is part of Aenea
#
# Aenea is free software: you can redistribute it and/or modify it under
# the terms of version 3 of the GNU Lesser General Public License as
# published by the Free Software Foundation.
#
# Aenea is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with Aenea.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright (2014) Alex Roper
# Alex Roper <alex@aroper.net>

"""
Mock module to allow dragonfly to be imported on linux locally.
Heavily modified to allow more dragonfly functionality to work
regardless of operating system.
"""
from .actions import ActionBase, DynStrActionBase


# Mock ActionBase and DynStrActionBase classes

def mock_action(*args, **kwargs):
    return ActionBase()


def mock_dyn_str_action(*args, **kwargs):
    return DynStrActionBase(*args, **kwargs)

Text = mock_dyn_str_action
Key = mock_dyn_str_action
Mouse = mock_dyn_str_action
Paste = mock_dyn_str_action
WaitWindow = mock_action
FocusWindow = mock_action
StartApp = mock_action
BringApp = mock_action
PlaySound = mock_action


class _WindowInfo(object):
    # TODO Use proxy contexts instead
    executable = ""
    title = ""
    handle = ""


class Window(object):
    @staticmethod
    def get_foreground():
        return _WindowInfo


class MockBase(object):
    def __init__(self, *args, **kwargs):
        pass


class HardwareInput(MockBase):
    pass


class Keyboard(MockBase):
    pass


class KeyboardInput(MockBase):
    pass


monitors = []


class Monitor(MockBase):
    pass


class MouseInput(MockBase):
    pass


class Typeable(object):
    pass

typeables = {}


def make_input_array(inputs):
    return inputs


def send_input_array(input_array):
    pass



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it 
#   under the terms of the GNU Lesser General Public License as published 
#   by the Free Software Foundation, either version 3 of the License, or 
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but 
#   WITHOUT ANY WARRANTY; without even the implied warranty of 
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public 
#   License along with Dragonfly.  If not, see 
#   <http://www.gnu.org/licenses/>.
#

import os.path
import re
from setuptools import setup, find_packages


#---------------------------------------------------------------------------
# Gather version from distribution file.

directory = os.path.abspath(os.path.dirname(__file__))
path = os.path.join(directory, "version.txt")
version_string = open(path).readline()
match = re.match(r"\s*(?P<rel>(?P<ver>\d+\.\d+)(?:\.\S+)*)\s*", version_string)
version = match.group("ver")
release = match.group("rel")


#---------------------------------------------------------------------------
# Set up package.

def read(*names):
    return open(os.path.join(os.path.dirname(__file__), *names)).read()


setup(
      name             = "dragonfly2",
      version          = release,
      description      = "Speech recognition extension library",
      author           = "Christo Butcher",
      author_email     = "dist.dragonfly@twizzy.biz",
      maintainer       = "Dane Finlay",
      maintainer_email = "Danesprite@gmail.com",
      license          = "LICENSE.txt",
      url              = "https://github.com/Danesprite/dragonfly",
      zip_safe         = False,  # To unzip documentation files.
      long_description = read("README.rst"),

      package_data={
                    "": ["*.txt", "*.rst"]
                   },

      install_requires=[
                        "setuptools >= 0.6c7",
                        "comtypes;platform_system=='Windows'",
                        "pywin32;platform_system=='Windows'",
                        "six",
                        "pyperclip >= 1.7.0",
                        "enum34;python_version<'3.4'",
                        "regex"
                       ],

      extras_require={
          "sphinx": [
                     "sphinxwrapper >= 1.1.1",
                     "pyjsgf >= 1.2.2",
                     "pyaudio"
                    ]
      },

      classifiers=[
                   "Environment :: Win32 (MS Windows)",
                   "Development Status :: 4 - Beta",
                   "License :: OSI Approved :: "
                   "GNU Library or Lesser General Public License (LGPL)",
                   "Operating System :: Microsoft :: Windows",
                   "Programming Language :: Python",
                  ],

      packages=find_packages(),

      test_suite="dragonfly.test.suites.natlink_suite",
     )



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it 
#   under the terms of the GNU Lesser General Public License as published 
#   by the Free Software Foundation, either version 3 of the License, or 
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but 
#   WITHOUT ANY WARRANTY; without even the implied warranty of 
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public 
#   License along with Dragonfly.  If not, see 
#   <http://www.gnu.org/licenses/>.
#

import unittest
import doctest
import pkg_resources
import os.path
from dragonfly.test.engine_suite import EngineTestSuite


#===========================================================================

common_names   = [
                  ".test_log",
                  ".test_actions",
                  ".test_parser",
#                  ".test_engine",
                  ".test_engine_nonexistent",
                  ".test_window",
                  ".test_timer",
                  ".test_dictation",
                  ".test_language_en_number",
                  ".test_language_de_number",
                  ".test_language_nl_number",
                  ".test_accessibility",
                  "doc:documentation/test_action_base_doctest.txt",
                  "doc:documentation/test_grammar_elements_basic_doctest.txt",
                  "doc:documentation/test_grammar_elements_compound_doctest.txt",
                  "doc:documentation/test_grammar_list_doctest.txt",
                 ]

natlink_names  = [
                  ".test_engine_natlink",
                  "doc:documentation/test_recobs_doctest.txt",
#                  "doc:documentation/test_word_formatting_v10_doctest.txt",
                  "doc:documentation/test_word_formatting_v11_doctest.txt",
                 ]

sapi5_names    = [
                  ".test_engine_sapi5",
                 ]

sphinx_names   = [
                  ".test_engine_sphinx",
                 ]
#===========================================================================

def build_suite(suite, names):
    # Determine the root directory of the source code files.  This is
    #  used for finding doctest files specified relative to that root.
    project_root = os.path.join(os.path.dirname(__file__), "..", "..")
    project_root = os.path.abspath(project_root)

    # Load test cases from specified names.
    loader = unittest.defaultTestLoader
    for name in names:
        if name.startswith("."):
            name = "dragonfly.test" + name
            suite.addTests(loader.loadTestsFromName(name))
        elif name.startswith("doc:"):
            path = name[4:]
            path = os.path.join(project_root, *path.split("/"))
            path = os.path.abspath(path)
            suite.addTests(doctest.DocFileSuite(path))
        else:
            raise Exception("Invalid test name: %r." % (name,))
    return suite


natlink_suite  = build_suite(EngineTestSuite("natlink"),
                             natlink_names + common_names)
sapi5_suite    = build_suite(EngineTestSuite("sapi5"),
                             sapi5_names + common_names)
sphinx_suite    = build_suite(EngineTestSuite("sphinx"),
                             sphinx_names + common_names)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it 
#   under the terms of the GNU Lesser General Public License as published 
#   by the Free Software Foundation, either version 3 of the License, or 
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but 
#   WITHOUT ANY WARRANTY; without even the implied warranty of 
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public 
#   License along with Dragonfly.  If not, see 
#   <http://www.gnu.org/licenses/>.
#

"""
Compound element classes
============================================================================

This file implements the Compound grammar element class for
creating grammar element structures based on a simple text format.

"""


import re
from six import string_types

import dragonfly.grammar.elements_basic as elements_
import dragonfly.parser as parser_
import logging


class _Stuff(parser_.Sequence):

    def __init__(self):
        super(_Stuff, self).__init__()
        self._single = None

    def initialize(self):
        self._single = _Single()
        self._single.initialize()
        repetition = parser_.Repetition(self._single)
        elements = (
            repetition,
            parser_.Repetition(
                parser_.Sequence((
                    parser_.String("|"),
                    repetition,
                    )),
                min=0),
            )
        parser_.Sequence.__init__(self, elements)

    def set_elements(self, identifiers):
        self._single.set_elements(identifiers)

    def set_actions(self, identifiers):
        self._single.set_actions(identifiers)

    def value(self, node):
        repetitions = [node.children[0].value()]
        for repetition in node.children[1].children:
            repetitions.append(repetition.children[1].value())

        alternatives = []
        for repetition in repetitions:
            if len(repetition) == 1:
                alternatives.append(repetition[0])
            else:
                element = elements_.Sequence(repetition)
                alternatives.append(element)

        if len(alternatives) == 1:
            return alternatives[0]
        else:
            return elements_.Alternative(alternatives)

stuff = _Stuff()


class _Single(parser_.Sequence):

    def __init__(self):
        super(_Single, self).__init__()
        self._action_identifier = None
        self._element_identifier = None

    def initialize(self):
        self._element_identifier = _ElementRef()
        self._action_identifier = _ActionRef()
        alternatives = parser_.Alternative((
            _Literal(),
            _Optional(),
            _Group(),
            self._element_identifier,
            ))
        elements = [
            parser_.Whitespace(optional=True),
            alternatives,
            parser_.Optional(
                parser_.Sequence((
                    parser_.Whitespace(optional=True),
                    self._action_identifier,
                    ))
                ),
            parser_.Whitespace(optional=True),
            ]
        parser_.Sequence.__init__(self, elements)

    def set_elements(self, identifiers):
        self._element_identifier.set_identifiers(identifiers)

    def set_actions(self, identifiers):
        self._action_identifier.set_identifiers(identifiers)

    def value(self, node):
        alternative = node.children[1]
        element = alternative.children[0].value()

        optional = node.children[2]
        if optional.children:
            action = optional.children[0].children[1].value()
            element.add_action(action)

        return element


class _ElementRef(parser_.Sequence):

    def __init__(self):
        # Use a pattern to allow ascii and Unicode alphanumeric characters plus
        # underscores.
        pattern = re.compile(r"\w", re.UNICODE)
        name = parser_.CharacterSeries(None, pattern=pattern)
        elements = (parser_.String("<"), name, parser_.String(">"))
        parser_.Sequence.__init__(self, elements)
        self._identifiers = None

    def set_identifiers(self, identifiers):
        self._identifiers = identifiers

    def value(self, node):
        name = node.children[1].value()
        try:
            return self._identifiers[name]
        except KeyError:
            root = node
            while root.parent:
                root = root.parent
            s = "Unknown reference name %r in %r" % (node.data, root.data)
            raise Exception(s)


class _ActionRef(parser_.Sequence):

    def __init__(self):
        # Use a pattern to allow ascii and Unicode alphanumeric characters plus
        # underscores.
        pattern = re.compile(r"\w", re.UNICODE)
        name = parser_.CharacterSeries(None, pattern=pattern)
        elements = (parser_.String("{"), name, parser_.String("}"))
        parser_.Sequence.__init__(self, elements)
        self._identifiers = None

    def set_identifiers(self, identifiers):
        self._identifiers = identifiers

    def value(self, node):
        name = node.children[1].value()
        try:
            return self._identifiers[name]
        except KeyError:
            root = node
            while root.parent:
                root = root.parent
            s = "Unknown reference name %r in %r" % (node.data, root.data)
            raise Exception(s)


class _Literal(parser_.Sequence):

    def __init__(self):
        # Use a pattern to allow ascii and Unicode alphanumeric characters plus a
        # few special characters.
        pattern = re.compile(r"[\w_\-.',]", re.UNICODE)
        word = parser_.CharacterSeries(None, pattern=pattern)
        whitespace = parser_.Whitespace()
        elements = (
            word,
            parser_.Repetition(
                parser_.Sequence((whitespace, word)),
                min=0),
            )
        parser_.Sequence.__init__(self, elements)

    def value(self, node):
        return elements_.Literal(node.match())


class _Optional(parser_.Sequence):

    def __init__(self):
        elements = (parser_.String("["), stuff, parser_.String("]"))
        parser_.Sequence.__init__(self, elements)

    def value(self, node):
        child = node.children[1].value()
        return elements_.Optional(child)


class _Group(parser_.Sequence):

    def __init__(self):
        elements = (parser_.String("("), stuff, parser_.String(")"))
        parser_.Sequence.__init__(self, elements)

    def value(self, node):
        child = node.children[1].value()
        return child


stuff.initialize()


#---------------------------------------------------------------------------
# The Compound class.

class Compound(elements_.Alternative):

    _log = logging.getLogger("compound.parse")
    _parser = parser_.Parser(stuff, _log)

    def __init__(self, spec, extras=None, actions=None, name=None,
                 value=None, value_func=None, elements=None, default=None):
        self._spec = spec
        self._value = value
        self._value_func = value_func

        if extras   is None:   extras   = {}
        if actions  is None:   actions  = {}
        if elements is None:   elements = {}

        # Convert extras argument from sequence to mapping.
        if isinstance(extras, (tuple, list)):
            mapping = {}
            for element in extras:
                if not isinstance(element, elements_.ElementBase):
                    self._log.error("Invalid extras item: %s" % element)
                    raise TypeError("Invalid extras item: %s" % element)
                if not element.name:
                    self._log.error("Extras item does not have a name: %s" % element)
                    raise TypeError("Extras item does not have a name: %s" % element)
                if element.name in mapping:
                    self._log.warning("Multiple extras items with the same name: %s" % element)
                mapping[element.name] = element
            extras = mapping
        elif not isinstance(extras, dict):
            self._log.error("Invalid extras argument: %s" % extras)
            raise TypeError("Invalid extras argument: %s" % extras)

        # Temporary transition code so that both "elements" and "extras"
        #  are supported as keyword arguments.
        if extras and elements:
            extras = dict(extras)
            extras.update(elements)
        elif elements:
            extras = elements
        self._extras = extras

        # This solution is non-ideal as "stuff" is a global instance.
        stuff.set_elements(extras)
        stuff.set_actions(actions)

        element = self._parser.parse(spec)
        if not element:
            self._log.error("Invalid compound spec: %r" % spec)
            raise SyntaxError("Invalid compound spec: %r" % spec)
        elements_.Alternative.__init__(self, (element,), name=name,
                                       default=default)

    def __str__(self):
        arguments = ["%r" % self._spec]
        if self.name:
            arguments.append("name=%r" % self.name)
        arguments = ", ".join(arguments)
        return "%s(%s)" % (self.__class__.__name__, arguments)

    def value(self, node):
        if self._value_func is not None:
            # Prepare *extras* dict for passing to value_func().
            extras = {"_node": node}
            for name, element in self._extras.items():
                extra_node = node.get_child_by_name(name, shallow=True)
                if extra_node:
                    extras[name] = extra_node.value()
                elif element.has_default():
                    extras[name] = element.default
            try:
                value = self._value_func(node, extras)
            except Exception as e:
                self._log.warning("Exception from value_func: %s" % e)
                raise
            return value
        elif self._value is not None:
            return self._value
        else:
            return elements_.Alternative.value(self, node)


#---------------------------------------------------------------------------
# The Choice class which maps multiple Compound instances to values.

class Choice(elements_.Alternative):

    def __init__(self, name, choices, extras=None, default=None):

        # Argument type checking.
        assert isinstance(name, string_types) or name is None
        assert isinstance(choices, dict)
        for k, v in choices.items():
            assert isinstance(k, string_types)

        # Construct children from the given choice keys and values.
        self._choices = choices
        self._extras = extras
        children = []
        for k, v in choices.items():
            child = Compound(spec=k, value=v, extras=extras)
            children.append(child)

        # Initialize super class.
        elements_.Alternative.__init__(self, children=children,
                                       name=name, default=default)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it 
#   under the terms of the GNU Lesser General Public License as published 
#   by the Free Software Foundation, either version 3 of the License, or 
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but 
#   WITHOUT ANY WARRANTY; without even the implied warranty of 
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public 
#   License along with Dragonfly.  If not, see 
#   <http://www.gnu.org/licenses/>.
#

"""
Dictation container class for the Sphinx engine
============================================================================

This class is used to store the recognized results of dictation elements 
within voice-commands.  It offers access to both the raw spoken-form words 
and be formatted written-form text.

The formatted text can be retrieved using
:meth:`~DictationContainerBase.format` or simply by  calling ``str(...)``
on a dictation container object. A tuple of the raw  spoken words can be
retrieved using :attr:`~DictationContainerBase.words`.

"""
from six import PY2

from ..base import DictationContainerBase


class SphinxDictationContainer(DictationContainerBase):
    """
        Container class for dictated words as recognized by the
        :class:`Dictation` element.

    """

    def __init__(self, words):
        DictationContainerBase.__init__(self, words=words)

    def __repr__(self):
        message = u"%s(%s)" % (self.__class__.__name__,
                               u", ".join(self._words))
        if PY2:
            return message.encode()
        else:
            return message



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it 
#   under the terms of the GNU Lesser General Public License as published 
#   by the Free Software Foundation, either version 3 of the License, or 
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but 
#   WITHOUT ANY WARRANTY; without even the implied warranty of 
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public 
#   License along with Dragonfly.  If not, see 
#   <http://www.gnu.org/licenses/>.
#

"""
BringApp and StartApp actions
============================================================================

The :class:`StartApp` and :class:`BringApp` action classes are used to 
start an application and bring it to the foreground.  :class:`StartApp` 
starts an application by running an executable file, while 
:class:`BringApp` first checks whether the application is already running 
and if so brings it to the foreground, otherwise starts it by running the 
executable file.


Example usage
----------------------------------------------------------------------------

The following example brings Notepad to the foreground if it is already 
open, otherwise it starts Notepad::

   BringApp(r"C:\Windows\system32\\notepad.exe").execute()

Note that the path to *notepad.exe* given above might not be correct for 
your computer, since it depends on the operating system and its 
configuration.

In some cases an application might be accessible simply through the file 
name of its executable, without specifying the directory.  This depends on 
the operating system's path configuration.  For example, on the author's 
computer the following command successfully starts Notepad::

   BringApp("notepad").execute()


Class reference
----------------------------------------------------------------------------

"""

import os.path
from subprocess           import Popen
from .action_base         import ActionBase, ActionError
from .action_focuswindow  import FocusWindow


#---------------------------------------------------------------------------

class StartApp(ActionBase):
    """
        Start an application.

        When this action is executed, it runs a file (executable), 
        optionally with commandline arguments.

    """

    def __init__(self, *args, **kwargs):
        """
            Constructor arguments:
             - *args* (variable argument list of *str*'s) --
               these strings are passed to subprocess.Popen()
               to start the application as a child process
             - *cwd* (*str*, default *None*) --
               if not *None*, then start the application in this
               directory
    
        """
        ActionBase.__init__(self)
        self._args = args

        if "cwd" in kwargs:  self._cwd = kwargs.pop("cwd")
        else:                self._cwd = None
        if kwargs:
            raise ArgumentError("Invalid keyword arguments: %r" % kwargs)

        # Expand any variables within path names.
        self._args = [self._interpret(a) for a in self._args]
        if self._cwd:
            self._cwd = self._interpret(self._cwd)

        self._str = str(", ".join(repr(a) for a in self._args))

    def _interpret(self, path):
        return os.path.expanduser(os.path.expandvars(path))

    def _execute(self, data=None):
        self._log.debug("Starting app: %r" % (self._args,))
        try:
            Popen(self._args, cwd=self._cwd)
        except Exception as e:
            raise ActionError("Failed to start app %s: %s" % (self._str, e))


#---------------------------------------------------------------------------

class BringApp(StartApp):
    """
        Bring an application to the foreground, starting it if it is not 
        yet running.

        When this action is executed, it looks for an existing window of 
        the application specified in the constructor arguments.  If an 
        existing window is found, that window is brought to the 
        foreground.  On the other hand, if no window is found the 
        application is started.

        Note that the constructor arguments are identical to those used by 
        the :class:`StartApp` action class.

    """

    def __init__(self, *args, **kwargs):
        """
            Constructor arguments:
             - *args* (variable argument list of *str*'s) --
               these strings are passed to :meth:`subprocess.Popen`
               to start the application as a child process
             - *cwd* (*str*, default *None*) --
               if not *None*, then start the application in this
               directory
    
        """
        StartApp.__init__(self, *args, **kwargs)

    def _execute(self, data=None):
        self._log.debug("Bringing app: %r" % (self._args,))
        target = self._args[0].lower()
        focus_action = FocusWindow(executable=target)
        # Attempt to focus on an existing window.
        if not focus_action.execute():
            # Failed to focus on an existing window, so start
            #  the application.
            StartApp._execute(self, data)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it 
#   under the terms of the GNU Lesser General Public License as published 
#   by the Free Software Foundation, either version 3 of the License, or 
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but 
#   WITHOUT ANY WARRANTY; without even the implied warranty of 
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public 
#   License along with Dragonfly.  If not, see 
#   <http://www.gnu.org/licenses/>.
#

"""
    This file implements a basic multiplexing interface to the natlink timer.
"""


import time
import logging

# Optionally import natlink
try:
    import natlink
except ImportError:
    natlink = None


class _Timer(object):

    class Callback(object):
        def __init__(self, function, interval):
            self.function = function
            self.interval = interval
            self.next_time = time.clock() + self.interval

        def call(self):
            self.next_time += self.interval
            try:
                self.function()
            except Exception as e:
                logging.getLogger("timer").exception("Exception during timer callback")
                print("Exception during timer callback: %s (%r)" % (e, e))

    def __init__(self, interval):
        self.interval = interval
        self.callbacks = []

    def add_callback(self, function, interval):
        self.callbacks.append(self.Callback(function, interval))
        if len(self.callbacks) == 1 and natlink:
            natlink.setTimerCallback(self.callback, int(self.interval * 1000))

    def remove_callback(self, function):
        for c in self.callbacks:
            if c.function == function:
                self.callbacks.remove(c)
        if len(self.callbacks) == 0 and natlink:
            natlink.setTimerCallback(None, 0)

    def callback(self):
        now = time.clock()
        for c in self.callbacks:
            if c.next_time < now:
                c.call()


timer = _Timer(0.025)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it 
#   under the terms of the GNU Lesser General Public License as published 
#   by the Free Software Foundation, either version 3 of the License, or 
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but 
#   WITHOUT ANY WARRANTY; without even the implied warranty of 
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public 
#   License along with Dragonfly.  If not, see 
#   <http://www.gnu.org/licenses/>.
#

"""
This file offers access to various action classes.

"""
import sys

# Import OS-agnostic classes
from .action_pause        import Pause
from .action_function     import Function
from .action_playback     import Playback
from .action_base         import (ActionBase, DynStrActionBase,
                                  Repeat, ActionError)
from .action_mimic        import Mimic
from .action_cmd          import RunCommand

# Import Windows OS dependent classes only for Windows
if sys.platform.startswith("win"):
    from .action_key          import Key
    from .action_text         import Text
    from .action_mouse        import Mouse
    from .action_paste        import Paste
    from .action_waitwindow   import WaitWindow
    from .action_focuswindow  import FocusWindow
    from .action_startapp     import StartApp, BringApp
    from .action_playsound    import PlaySound
    from .keyboard             import Keyboard, Typeable
else:
    from ..os_dependent_mock import Key
    from ..os_dependent_mock import Text
    from ..os_dependent_mock import Mouse
    from ..os_dependent_mock import Paste
    from ..os_dependent_mock import WaitWindow
    from ..os_dependent_mock import FocusWindow
    from ..os_dependent_mock import StartApp, BringApp
    from ..os_dependent_mock import PlaySound
    from ..os_dependent_mock import Keyboard, Typeable
